# Generated by Django 4.2.20 on 2025-05-22 15:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0237_enterprisecustomersupportusersview'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_learner_credit_message_box',
            field=models.BooleanField(default=True, help_text='Specifies whether to display the Learner Credit message box in the learner portal.', verbose_name='Display learner credit message box in learner portal'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_learner_credit_message_box',
            field=models.BooleanField(default=True, help_text='Specifies whether to display the Learner Credit message box in the learner portal.', verbose_name='Display learner credit message box in learner portal'),
        ),
    ]



"""
    Serializer for Blackboard configuration.
"""
from rest_framework import serializers

from integrated_channels.api.serializers import EnterpriseCustomerPluginConfigSerializer
from integrated_channels.blackboard.models import (
    BlackboardEnterpriseCustomerConfiguration,
    BlackboardGlobalConfiguration,
)


class BlackboardConfigSerializer(EnterpriseCustomerPluginConfigSerializer):
    oauth_authorization_url = serializers.ReadOnlyField()

    class Meta:
        model = BlackboardEnterpriseCustomerConfiguration
        extra_fields = (
            'encrypted_client_id',
            'encrypted_client_secret',
            'blackboard_base_url',
            'refresh_token',
            'uuid',
            'oauth_authorization_url',
        )
        fields = EnterpriseCustomerPluginConfigSerializer.Meta.fields + extra_fields

    encrypted_client_id = serializers.CharField(required=False, allow_blank=False, read_only=False)
    encrypted_client_secret = serializers.CharField(required=False, allow_blank=False, read_only=False)


class BlackboardGlobalConfigSerializer(serializers.ModelSerializer):

    class Meta:
        model = BlackboardGlobalConfiguration
        fields = ('app_key',)



# Generated by Django 4.2.10 on 2024-06-14 08:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0214_alter_enterprisecustomer_learner_portal_sidebar_content_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='enterprisecustomer',
            name='career_engagement_network_message',
        ),
        migrations.RemoveField(
            model_name='enterprisecustomer',
            name='enable_career_engagement_network_on_learner_portal',
        ),
        migrations.RemoveField(
            model_name='historicalenterprisecustomer',
            name='career_engagement_network_message',
        ),
        migrations.RemoveField(
            model_name='historicalenterprisecustomer',
            name='enable_career_engagement_network_on_learner_portal',
        ),
    ]



# Generated by Django 4.2.13 on 2024-05-29 07:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('xapi', '0011_alter_xapilearnerdatatransmissionaudit_index_together'),
    ]

    operations = [
        migrations.AddField(
            model_name='xapilearnerdatatransmissionaudit',
            name='is_transmitted',
            field=models.BooleanField(default=False),
        ),
    ]



"""
Client for connecting to Degreed.
"""

import datetime
import logging
import time
from urllib.parse import urljoin

import requests

from django.apps import apps

from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.utils import generate_formatted_log

LOGGER = logging.getLogger(__name__)


class DegreedAPIClient(IntegratedChannelApiClient):
    """
    Client for connecting to Degreed.

    Specifically, this class supports obtaining access tokens and posting to the courses and
    completion status endpoints.
    """

    CONTENT_PROVIDER_SCOPE = 'provider_content'
    COMPLETION_PROVIDER_SCOPE = 'provider_completion'
    SESSION_TIMEOUT = 60

    def __init__(self, enterprise_configuration):
        """
        Instantiate a new client.

        Args:
            enterprise_configuration (DegreedEnterpriseCustomerConfiguration): An enterprise customers's
            configuration model for connecting with Degreed
        """
        super().__init__(enterprise_configuration)
        self.global_degreed_config = apps.get_model('degreed', 'DegreedGlobalConfiguration').current()
        self.session = None
        self.expires_at = None

    def create_assessment_reporting(self, user_id, payload):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Degreed integrated channel does not yet support assessment reporting."
            )
        )

    def cleanup_duplicate_assignment_records(self, courses):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Degreed integrated channel does not yet support assignment deduplication."
            )
        )

    def create_course_completion(self, user_id, payload):
        """
        Send a completion status payload to the Degreed Completion Status endpoint

        Args:
            user_id: Unused.
            payload: JSON encoded object (serialized from DegreedLearnerDataTransmissionAudit)
                containing completion status fields per Degreed documentation.

        Returns:
            A tuple containing the status code and the body of the response.
        Raises:
            HTTPError: if we received a failure response code from Degreed
        """
        return self._post(
            urljoin(
                self.enterprise_configuration.degreed_base_url,
                self.global_degreed_config.completion_status_api_path
            ),
            payload,
            self.COMPLETION_PROVIDER_SCOPE
        )

    def delete_course_completion(self, user_id, payload):
        """
        Delete a completion status previously sent to the Degreed Completion Status endpoint

        Args:
            user_id: Unused.
            payload: JSON encoded object (serialized from DegreedLearnerDataTransmissionAudit)
                containing the required completion status fields for deletion per Degreed documentation.

        Returns:
            A tuple containing the status code and the body of the response.
        Raises:
            HTTPError: if we received a failure response code from Degreed
        """
        return self._delete(
            urljoin(
                self.enterprise_configuration.degreed_base_url,
                self.global_degreed_config.completion_status_api_path
            ),
            payload,
            self.COMPLETION_PROVIDER_SCOPE
        )

    def create_content_metadata(self, serialized_data):
        """
        Create content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        self._sync_content_metadata(serialized_data, 'post')

    def update_content_metadata(self, serialized_data):
        """
        Update content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        self._sync_content_metadata(serialized_data, 'post')

    def delete_content_metadata(self, serialized_data):
        """
        Delete content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        self._sync_content_metadata(serialized_data, 'delete')

    def _sync_content_metadata(self, serialized_data, http_method):
        """
        Synchronize content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.
            http_method: The HTTP method to use for the API request.

        Raises:
            ClientError: If Degreed API request fails.
        """
        try:
            status_code, response_body = getattr(self, '_' + http_method)(
                urljoin(self.enterprise_configuration.degreed_base_url, self.global_degreed_config.course_api_path),
                serialized_data,
                self.CONTENT_PROVIDER_SCOPE
            )
        except requests.exceptions.RequestException as exc:
            raise ClientError(
                'DegreedAPIClient request failed: {error} {message}'.format(
                    error=exc.__class__.__name__,
                    message=str(exc)
                )
            ) from exc

        if status_code >= 400:
            raise ClientError(
                'DegreedAPIClient request failed with status {status_code}: {message}'.format(
                    status_code=status_code,
                    message=response_body
                )
            )

    def _post(self, url, data, scope):
        """
        Make a POST request using the session object to a Degreed endpoint.

        Args:
            url (str): The url to send a POST request to.
            data (str): The json encoded payload to POST.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_PROVIDER_SCOPE`
                        - `COMPLETION_PROVIDER_SCOPE`
        """
        self._create_session(scope)
        response = self.session.post(url, data=data)
        return response.status_code, response.text

    def _delete(self, url, data, scope):
        """
        Make a DELETE request using the session object to a Degreed endpoint.

        Args:
            url (str): The url to send a DELETE request to.
            data (str): The json encoded payload to DELETE.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_PROVIDER_SCOPE`
                        - `COMPLETION_PROVIDER_SCOPE`
        """
        self._create_session(scope)
        response = self.session.delete(url, data=data)
        return response.status_code, response.text

    def _create_session(self, scope):
        """
        Instantiate a new session object for use in connecting with Degreed
        """
        now = datetime.datetime.utcnow()
        if self.session is None or self.expires_at is None or now >= self.expires_at:
            # Create a new session with a valid token
            if self.session:
                self.session.close()
            oauth_access_token, expires_at = self._get_oauth_access_token(
                self.enterprise_configuration.key,
                self.enterprise_configuration.secret,
                self.enterprise_configuration.degreed_user_id,
                self.enterprise_configuration.degreed_user_password,
                scope
            )
            session = requests.Session()
            session.timeout = self.SESSION_TIMEOUT
            session.headers['Authorization'] = 'Bearer {}'.format(oauth_access_token)
            session.headers['content-type'] = 'application/json'
            self.session = session
            self.expires_at = expires_at

    def _get_oauth_access_token(self, client_id, client_secret, user_id, user_password, scope):
        """ Retrieves OAuth 2.0 access token using the client credentials grant.

        Args:
            client_id (str): API client ID
            client_secret (str): API client secret
            user_id (str): Degreed company ID
            user_password (str): Degreed user password
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_PROVIDER_SCOPE`
                        - `COMPLETION_PROVIDER_SCOPE`

        Returns:
            tuple: Tuple containing access token string and expiration datetime.
        Raises:
            HTTPError: If we received a failure response code from Degreed.
            ClientError: If an unexpected response format was received that we could not parse.
        """
        response = requests.post(
            urljoin(self.enterprise_configuration.degreed_base_url, self.global_degreed_config.oauth_api_path),
            data={
                'grant_type': 'password',
                'username': user_id,
                'password': user_password,
                'scope': scope,
            },
            auth=(client_id, client_secret),
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )

        try:
            data = response.json()
            expires_at = data['expires_in'] + int(time.time())
            return data['access_token'], datetime.datetime.utcfromtimestamp(expires_at)
        except (KeyError, ValueError) as error:
            raise ClientError(response.text, response.status_code) from error



# Generated by Django 4.2.20 on 2025-05-05 19:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0236_create_vw_enterprise_customer_ecus_pecus'),
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseCustomerSupportUsersView',
            fields=[
                ('enterprise_customer_id', models.UUIDField()),
                ('user_email', models.EmailField(max_length=254, primary_key=True, serialize=False)),
                ('is_pending', models.BooleanField(default=False)),
                ('user_id', models.PositiveIntegerField()),
                ('username', models.CharField(max_length=150)),
                ('first_name', models.CharField(max_length=150)),
                ('last_name', models.CharField(max_length=150)),
                ('is_staff', models.BooleanField(default=False)),
                ('is_active', models.BooleanField(default=False)),
                ('date_joined', models.DateTimeField(null=True)),
                ('is_admin', models.BooleanField(default=False)),
            ],
            options={
                'db_table': 'view_enterprise_customer_support_users',
                'managed': False,
            },
        ),
    ]



# Generated by Django 3.2.23 on 2023-11-30 19:01

from django.db import migrations, models


def backfill_sso_marked_authorized(apps, app_schema):
    """
    Sets all Enterprise Customer SSO Configuration records 'marked_authorized' to False, in anticipation
    of making it the default value
    """
    EnterpriseCustomerSsoConfiguration = apps.get_model('enterprise', 'EnterpriseCustomerSsoConfiguration')
    queryset = EnterpriseCustomerSsoConfiguration.all_objects.all()
    for sso_config in queryset:
        sso_config.marked_authorized = False
    EnterpriseCustomerSsoConfiguration.all_objects.bulk_update(queryset, ['marked_authorized'])


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0195_auto_20231130_1837'),
    ]

    operations = [
        migrations.RunPython(
            code=backfill_sso_marked_authorized,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



"""
Tests for enterprise middleware.
"""

import unittest
from unittest import mock

import ddt
from pytest import mark

from django.conf import settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import Client, RequestFactory

from enterprise.middleware import EnterpriseLanguagePreferenceMiddleware
from test_utils.factories import (
    AnonymousUserFactory,
    EnterpriseCustomerFactory,
    EnterpriseCustomerUserFactory,
    UserFactory,
)

LANGUAGE_KEY = 'pref-lang'
UserAPIInternalError = Exception
UserAPIRequestError = Exception


@mark.django_db
@ddt.ddt
class TestUserPreferenceMiddleware(unittest.TestCase):
    """
    Tests to make sure language configured by the `default_language` column on EnterpriseCustomer is being used.
    """

    def setUp(self):
        """
        Setup middleware, request, session, user and enterprise customer for tests. Also mock imports from edx-platform.
        """
        self.mock_imports()

        super().setUp()
        self.mock_response = mock.Mock()
        self.middleware = EnterpriseLanguagePreferenceMiddleware(self.mock_response)
        self.session_middleware = SessionMiddleware(self.mock_response)
        self.user = UserFactory.create()
        self.anonymous_user = AnonymousUserFactory()
        self.request = RequestFactory().get('/somewhere')
        self.request.user = self.user
        self.request.META['HTTP_ACCEPT_LANGUAGE'] = 'ar;q=1.0'
        self.session_middleware.process_request(self.request)
        self.client = Client()
        self.enterprise_customer = EnterpriseCustomerFactory()
        self.enterprise_customer_user = EnterpriseCustomerUserFactory(
            enterprise_customer=self.enterprise_customer,
            user_id=self.user.id,
        )

    def mock_imports(self):
        """
        Mock all the imports from edx-platform
        """
        mocks = [
            mock.patch('enterprise.middleware.LANGUAGE_KEY', LANGUAGE_KEY),
            mock.patch('enterprise.middleware.UserAPIInternalError', UserAPIInternalError),
            mock.patch('enterprise.middleware.UserAPIRequestError', UserAPIRequestError),
            mock.patch('enterprise.middleware.get_user_preference', mock.MagicMock(return_value=None)),
            mock.patch('enterprise.middleware.is_request_from_mobile_app', mock.MagicMock(return_value=False)),
            mock.patch('enterprise.utils.UserPreference', mock.MagicMock()),

        ]
        for mock_object in mocks:
            self.jwt_builder = mock_object.start()
            self.addCleanup(mock_object.stop)

    @ddt.data(None, 'es', 'en')
    def test_preference_setting_changes_cookie(self, lang_pref_out):
        """
        Validate that the language set via enterprise customer's `default_language`
        column is used as the learner's default language.
        """
        self.enterprise_customer.default_language = lang_pref_out
        self.enterprise_customer.save()
        self.middleware.process_request(self.request)

        assert getattr(self.request, '_anonymous_user_cookie_lang', None) == lang_pref_out

    def test_real_user_extracted_from_request(self):
        """
        Validate the the real_user is used in cases where user is masquerading as someone else.
        """
        # Hide the real user and masquerade as a fake user, fake user does not belong to any enterprise customer.
        self.request.user = UserFactory()
        self.request.user.real_user = self.user

        self.middleware.process_request(self.request)

        # Make sure the real user is used for setting the language cookie
        assert getattr(self.request, '_anonymous_user_cookie_lang', None) == self.enterprise_customer.default_language

    def test_cookie_not_set_for_anonymous_user(self):
        """
        Validate the language cookie is not set if the request user is not authenticated.
        """
        # Hide the real user and masquerade as a fake user, fake user does not belong to any enterprise customer.
        self.request.user = self.anonymous_user
        self.middleware.process_request(self.request)

        # Make sure the set cookie is not called for anonymous users
        assert getattr(self.request, '_anonymous_user_cookie_lang', None) is None

    def test_cookie_not_set_for_non_enterprise_learners(self):
        """
        Validate the language cookie is not set if the request user does not belong to any enterprise customer.
        """
        # Hide the real user and masquerade as a fake user, fake user does not belong to any enterprise customer.
        self.request.user = UserFactory()
        self.middleware.process_request(self.request)

        # Make sure the set cookie is not called for anonymous users
        assert getattr(self.request, '_anonymous_user_cookie_lang', None) is None

    def test_cookie_when_there_is_no_request_user(self):
        """
        Validate the language cookie is not set if, for some reason, the request user is not present.
        """
        # Hide the real user and masquerade as a fake user, fake user does not belong to any enterprise customer.
        request = RequestFactory().get('/somewhere')
        self.mock_response = mock.Mock()
        session_middleware = SessionMiddleware(self.mock_response)
        session_middleware.process_request(request)

        self.middleware.process_request(request)

        # Make sure the set cookie is not called for anonymous users
        assert getattr(self.request, '_anonymous_user_cookie_lang', None) is None

    def test_errors_are_handled(self):
        """
        Validate that the errors raised when querying user preference are handled correctly.
        In this case those errors are ignored.
        """
        with mock.patch('enterprise.middleware.get_user_preference') as mock_get_user_preference:
            mock_get_user_preference.side_effect = UserAPIInternalError
            self.middleware.process_request(self.request)

            # Make sure the set cookie is not called for anonymous users
            # pylint: disable=protected-access,no-member
            assert self.request._anonymous_user_cookie_lang == self.enterprise_customer.default_language

    def test_cookie_not_set_for_mobile_requests(self):
        """
        Validate the language cookie is not set if the request is coming from the mobile app.
        """
        with mock.patch('enterprise.middleware.is_request_from_mobile_app') as mock_is_request_from_mobile_app:
            mock_is_request_from_mobile_app.return_value = True
            self.middleware.process_request(self.request)

            # Make sure the set cookie is not called for anonymous users
            assert getattr(self.request, '_anonymous_user_cookie_lang', None) is None

    def test_middleware_when_cookie_lang_is_different_from_user_pref(self):
        """
        Validate that when user pref and cookie language are set but have different values
        then the middleware updates the cookie with user preference value.
        """
        user_pref_lang = 'ar'
        cookie_lang = 'en'
        self.request.COOKIES[settings.LANGUAGE_COOKIE_NAME] = cookie_lang

        with mock.patch('enterprise.middleware.get_user_preference') as mock_get_user_preference:
            mock_get_user_preference.return_value = user_pref_lang
            self.middleware.process_request(self.request)

            assert self.request.COOKIES[settings.LANGUAGE_COOKIE_NAME] == user_pref_lang
            assert getattr(self.request, '_anonymous_user_cookie_lang', None) == user_pref_lang



"""
Remove content transmission audit records that do not contain a catalog UUID.
"""
import logging

from django.core.management.base import BaseCommand

from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin
from integrated_channels.integrated_channel.tasks import remove_null_catalog_transmission_audits

LOGGER = logging.getLogger(__name__)


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Remove content transmission audit records that do not contain a catalog UUID.
    ./manage.py lms remove_null_catalog_transmission_audits
    """

    def handle(self, *args, **options):
        """
        Filter content transmission audit records that do not contain a catalog UUID and remove them.
        """
        try:
            remove_null_catalog_transmission_audits.delay()
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.exception(
                f'''Failed to remove content transmission audits that do not
                contain a catalog UUID. Task failed with exception: {exc}'''
            )



"""
Tests for the `roles_api` module.
"""
from django.test import TestCase

from enterprise import roles_api
from enterprise.constants import (
    ENTERPRISE_ADMIN_ROLE,
    ENTERPRISE_LEARNER_ROLE,
    ENTERPRISE_OPERATOR_ROLE,
    SYSTEM_ENTERPRISE_CATALOG_ADMIN_ROLE,
    SYSTEM_ENTERPRISE_PROVISIONING_ADMIN_ROLE,
)
from enterprise.models import SystemWideEnterpriseRole


class TestUpdateRoleAssignmentsCommand(TestCase):
    """
    Tests the roles_api functions.
    """
    ALL_ROLE_NAMES = (
        ENTERPRISE_ADMIN_ROLE,
        ENTERPRISE_LEARNER_ROLE,
        ENTERPRISE_OPERATOR_ROLE,
        SYSTEM_ENTERPRISE_CATALOG_ADMIN_ROLE,
        SYSTEM_ENTERPRISE_PROVISIONING_ADMIN_ROLE
    )

    def setUp(self):
        """ Creates role objects for each system role specified in constants."""
        super().setUp()
        for role_name in self.ALL_ROLE_NAMES:
            SystemWideEnterpriseRole.objects.get_or_create(name=role_name)

    def tearDown(self):
        """ Delete any existing role objects."""
        super().tearDown()
        SystemWideEnterpriseRole.objects.all().delete()

    def test_roles_by_name(self):
        for role_name in self.ALL_ROLE_NAMES:
            role_object = roles_api.roles_by_name().get(role_name)
            self.assertEqual(role_name, role_object.name)



# Generated by Django 4.2.20 on 2025-05-22 15:18

from django.db import migrations


def backfill_enable_learner_credit_message_box(apps, schema_editor):
    """
    Backfill enable_learner_credit_message_box field for all existing EnterpriseCustomer records.
    """
    EnterpriseCustomer = apps.get_model('enterprise', 'EnterpriseCustomer')
    EnterpriseCustomer.objects.all().update(enable_learner_credit_message_box=True)

class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0238_enterprisecustomer_enable_learner_credit_message_box_and_more'),
    ]

    operations = [
        migrations.RunPython(
            backfill_enable_learner_credit_message_box,
            reverse_code=migrations.RunPython.noop
        ),
    ]



# Generated by Django 2.2.19 on 2021-03-01 13:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0123_enterprisecustomeridentityprovider_default_provider'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='sender_alias',
            field=models.CharField(blank=True, help_text='Specifies enterprise customized sender alias.', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='sender_alias',
            field=models.CharField(blank=True, help_text='Specifies enterprise customized sender alias.', max_length=255, null=True),
        ),
    ]



"""
URLs for enterprise.
"""

from edx_api_doc_tools import make_api_info, make_docs_urls

from django.conf import settings
from django.urls import include, path, re_path

from enterprise.constants import COURSE_KEY_URL_PATTERN
from enterprise.heartbeat.views import heartbeat
from enterprise.views import (
    EnterpriseLoginView,
    EnterpriseProxyLoginView,
    EnterpriseSelectionView,
    GrantDataSharingPermissions,
    RouterView,
)

ENTERPRISE_ROUTER = RouterView.as_view()

enterprise_rest_api_urls = path(
    'enterprise/api/',
    include('enterprise.api.urls'),
    name='enterprise_api'
)

urlpatterns = [
    re_path(r'^enterprise/grant_data_sharing_permissions', GrantDataSharingPermissions.as_view(),
            name='grant_data_sharing_permissions'
            ),
    re_path(r'^enterprise/select/active', EnterpriseSelectionView.as_view(),
            name='enterprise_select_active'
            ),
    re_path(r'^enterprise/login', EnterpriseLoginView.as_view(),
            name='enterprise_slug_login'
            ),
    re_path(r'^enterprise/proxy-login', EnterpriseProxyLoginView.as_view(),
            name='enterprise_proxy_login'
            ),
    re_path(
        r'^enterprise/handle_consent_enrollment/(?P<enterprise_uuid>[^/]+)/course/{}/$'.format(
            settings.COURSE_ID_PATTERN
        ),
        ENTERPRISE_ROUTER,
        name='enterprise_handle_consent_enrollment'
    ),
    re_path(
        r'^enterprise/(?P<enterprise_uuid>[^/]+)/course/{}/enroll/$'.format(COURSE_KEY_URL_PATTERN),
        ENTERPRISE_ROUTER,
        name='enterprise_course_enrollment_page'
    ),
    re_path(
        r'^enterprise/(?P<enterprise_uuid>[^/]+)/course/{}/enroll/$'.format(settings.COURSE_ID_PATTERN),
        ENTERPRISE_ROUTER,
        name='enterprise_course_run_enrollment_page'
    ),
    path(
        'enterprise/<uuid:enterprise_uuid>/program/<uuid:program_uuid>/enroll/',
        ENTERPRISE_ROUTER,
        name='enterprise_program_enrollment_page'
    ),
    re_path(
        r'^enterprise/heartbeat/', heartbeat,
        name='enterprise_heartbeat',
    ),
    enterprise_rest_api_urls,
]

# Because ROOT_URLCONF points here, we are including the urls from the other apps here for now.
urlpatterns += [
    path(
        '',
        include('consent.urls'),
        name='consent'
    ),
]

# When enabled, these URLs will be registered and take precedence over new plugin URLs
if getattr(settings, 'ENABLE_LEGACY_INTEGRATED_CHANNELS', True):
    urlpatterns += [
        path(
            'cornerstone/',
            include('integrated_channels.cornerstone.urls'),
            name='cornerstone'
        ),
        path(
            'canvas/',
            include('integrated_channels.canvas.urls'),
            name='canvas',
        ),
        path(
            'blackboard/',
            include('integrated_channels.blackboard.urls'),
            name='blackboard',
        ),
    ]

urlpatterns += [
    path(
        'enterprise_learner_portal/',
        include('enterprise_learner_portal.urls'),
        name='enterprise_learner_portal_api'
    ),
    path(
        'integrated_channels/api/',
        include('integrated_channels.api.urls')
    ),
]

api_docs_urlpatterns = make_docs_urls(
    make_api_info(
        title='Enterprise API',
        version='v1',
        description='Docs for the edx-enterprise `/enterprise/api/v1` REST API.',
    ),
    api_url_patterns=[enterprise_rest_api_urls],
)

urlpatterns.append(
    path('enterprise/', include(api_docs_urlpatterns)),
)



"""
Views for the ``enterprise-catalog-query`` API endpoint.
"""

from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import permissions
from rest_framework.authentication import SessionAuthentication
from rest_framework.pagination import PageNumberPagination

from enterprise import models
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadOnlyModelViewSet


class ExpandDefaultPageSize(PageNumberPagination):
    """
    Expands page size for the API.
    Used to populate support-tools repo's provisioning form catalog query dropdown component.
    """
    page_size = 100


class EnterpriseCatalogQueryViewSet(EnterpriseReadOnlyModelViewSet):
    """
    API views for the ``enterprise_catalog_query`` API endpoint.
    """
    queryset = models.EnterpriseCatalogQuery.objects.all()
    serializer_class = serializers.EnterpriseCatalogQuerySerializer
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser,)
    authentication_classes = (JwtAuthentication, SessionAuthentication,)
    pagination_class = ExpandDefaultPageSize



"""
URL definitions for Blackboard API.
"""

from django.urls import path

from integrated_channels.blackboard.views import BlackboardCompleteOAuthView

urlpatterns = [
    path('oauth-complete', BlackboardCompleteOAuthView.as_view(),
         name='blackboard-oauth-complete'
         ),
]



# Generated migration for MariaDB UUID field conversion (Django 5.2)
"""
Migration to convert UUIDField from char(32) to uuid type for MariaDB compatibility.

See: https://www.albertyw.com/note/django-5-mariadb-uuidfield
"""

from django.db import migrations


def apply_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    
    if connection.vendor != 'mysql':
        return
    
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    
    with connection.cursor() as cursor:
        # EnterpriseCustomer
        cursor.execute("ALTER TABLE enterprise_enterprisecustomer MODIFY uuid uuid NOT NULL")
        
        # PendingEnrollment
        cursor.execute("ALTER TABLE enterprise_pendingenrollment MODIFY license_uuid uuid NULL")
        cursor.execute("ALTER TABLE enterprise_historicalpendingenrollment MODIFY license_uuid uuid NULL")
        
        # EnterpriseCourseEntitlement
        cursor.execute("ALTER TABLE enterprise_enterprisecourseentitlement MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisecourseentitlement MODIFY uuid uuid NOT NULL")
        
        # Other enterprise models with UUIDs
        cursor.execute("ALTER TABLE enterprise_learnercreditenterprisecourseenrollment MODIFY transaction_id uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_licensedenterprisecourseenrollment MODIFY license_uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_defaultenterpriseenrollmentintention MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecatalogquery MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomercatalog MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerreportingconfiguration MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerinvitekey MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisecustomerinvitekey MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_chatgptresponse MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerssoconfiguration MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisegroup MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisegroup MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisegroupmembership MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_onboardingflow MODIFY uuid uuid NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomeradmin MODIFY uuid uuid NOT NULL")


def reverse_mariadb_migration(apps, schema_editor):
    connection = schema_editor.connection
    
    if connection.vendor != 'mysql':
        return
    
    with connection.cursor() as cursor:
        cursor.execute("SELECT VERSION()")
        version = cursor.fetchone()[0]
        if 'mariadb' not in version.lower():
            return
    
    with connection.cursor() as cursor:
        cursor.execute("ALTER TABLE enterprise_enterprisecustomer MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_pendingenrollment MODIFY license_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_historicalpendingenrollment MODIFY license_uuid char(32) NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecourseentitlement MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisecourseentitlement MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_learnercreditenterprisecourseenrollment MODIFY transaction_id char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_licensedenterprisecourseenrollment MODIFY license_uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_defaultenterpriseenrollmentintention MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecatalogquery MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomercatalog MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerreportingconfiguration MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerinvitekey MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisecustomerinvitekey MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_chatgptresponse MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomerssoconfiguration MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisegroup MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_historicalenterprisegroup MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisegroupmembership MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_onboardingflow MODIFY uuid char(32) NOT NULL")
        cursor.execute("ALTER TABLE enterprise_enterprisecustomeradmin MODIFY uuid char(32) NOT NULL")


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0241_alter_defaultenterpriseenrollmentintention_realized_enrollments'),
    ]

    operations = [
        migrations.RunPython(
            code=apply_mariadb_migration,
            reverse_code=reverse_mariadb_migration,
        ),
    ]



"""
Django signal handlers for integrated channels user retirement.
"""
from logging import getLogger

from enterprise.models import EnterpriseCourseEnrollment, EnterpriseCustomerUser
from integrated_channels.degreed.models import DegreedLearnerDataTransmissionAudit
from integrated_channels.sap_success_factors.models import SapSuccessFactorsLearnerDataTransmissionAudit

logger = getLogger(__name__)

try:
    from openedx.core.djangoapps.user_api.accounts.signals import USER_RETIRE_LMS_CRITICAL
except ImportError:
    USER_RETIRE_LMS_CRITICAL = None


def retire_sapsf_data_transmission(sender, user, **kwargs):  # pylint: disable=unused-argument
    """
    Handle USER_RETIRE_LMS_CRITICAL signal: clear sapsf_user_id on audit records.

    Idempotent: only updates records where sapsf_user_id is not already empty.
    """
    logger.info(f'Retiring sapsf_user_id {user.id} on audit records')
    for ent_user in EnterpriseCustomerUser.objects.filter(user_id=user.id):
        for enrollment in EnterpriseCourseEnrollment.objects.filter(enterprise_customer_user=ent_user):
            logger.info(f'Updating enrollment {enrollment.id} to clear sapsf_user_id {user.id}')
            SapSuccessFactorsLearnerDataTransmissionAudit.objects.filter(
                enterprise_course_enrollment_id=enrollment.id,
            ).exclude(
                sapsf_user_id="",
            ).update(sapsf_user_id="")


def retire_degreed_data_transmission(sender, user, **kwargs):  # pylint: disable=unused-argument
    """
    Handle USER_RETIRE_LMS_CRITICAL signal: clear degreed_user_email on audit records.

    Idempotent: only updates records where degreed_user_email is not already empty.
    """
    logger.info(f'Retiring degreed_user_email on audit records for user {user.id}')
    for ent_user in EnterpriseCustomerUser.objects.filter(user_id=user.id):
        for enrollment in EnterpriseCourseEnrollment.objects.filter(enterprise_customer_user=ent_user):
            logger.info(f'Updating enrollment {enrollment.id} to clear degreed_user_email for ECU {ent_user}')
            DegreedLearnerDataTransmissionAudit.objects.filter(
                enterprise_course_enrollment_id=enrollment.id,
            ).exclude(
                degreed_user_email="",
            ).update(degreed_user_email="")


if USER_RETIRE_LMS_CRITICAL is not None:
    USER_RETIRE_LMS_CRITICAL.connect(retire_sapsf_data_transmission)
    USER_RETIRE_LMS_CRITICAL.connect(retire_degreed_data_transmission)



"""
Errors thrown by the APIs in the Consent application.
"""


class ConsentAPIRequestError(Exception):
    """There was a problem with a request to the Consent application's APIs."""


class ConsentAPIInternalError(Exception):
    """An internal error occurred in one of the Consent application's APIs."""


class ConsentNotProvided(ConsentAPIRequestError):
    """Consent related to this request was not provided."""


class DataSharingConsentAPIError(ConsentAPIRequestError):
    """There was an error with a request to the DSC API."""


class InvalidProxyConsent(Exception):
    """A proxy consent object with the given details could not be created."""



# Generated by Django 3.2.23 on 2024-02-13 20:40

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0199_auto_20240130_0628'),
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseGroup',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(help_text='Specifies enterprise group name.', max_length=25)),
                ('enterprise_customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='enterprise.enterprisecustomer')),
            ],
            options={
                'verbose_name': 'Enterprise Group',
                'verbose_name_plural': 'Enterprise Groups',
                'ordering': ['-modified'],
                'unique_together': {('name', 'enterprise_customer')},
            },
        ),
        migrations.CreateModel(
            name='HistoricalEnterpriseGroupMembership',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('enterprise_customer_user', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.enterprisecustomeruser')),
                ('group', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.enterprisegroup')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('pending_enterprise_customer_user', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.pendingenterprisecustomeruser')),
            ],
            options={
                'verbose_name': 'historical Enterprise Group Membership',
                'verbose_name_plural': 'historical Enterprise Group Memberships',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': ('history_date', 'history_id'),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalEnterpriseGroup',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('name', models.CharField(help_text='Specifies enterprise group name.', max_length=25)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('enterprise_customer', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.enterprisecustomer')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical Enterprise Group',
                'verbose_name_plural': 'historical Enterprise Groups',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': ('history_date', 'history_id'),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='EnterpriseGroupMembership',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('enterprise_customer_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='enterprise.enterprisecustomeruser')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='members', to='enterprise.enterprisegroup')),
                ('pending_enterprise_customer_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='enterprise.pendingenterprisecustomeruser')),
            ],
            options={
                'verbose_name': 'Enterprise Group Membership',
                'verbose_name_plural': 'Enterprise Group Memberships',
                'ordering': ['-modified'],
                'unique_together': {('group', 'pending_enterprise_customer_user'), ('group', 'enterprise_customer_user')},
            },
        ),
    ]




"""
Views for `EnterpriseCustomerAdmin` model.
"""
import logging

from edx_rbac.decorators import permission_required
from rest_framework import mixins, status, viewsets
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import DatabaseError, transaction
from django.shortcuts import get_object_or_404

from enterprise import models, roles_api
from enterprise.api import utils as admin_utils
from enterprise.api.v1.serializers import AdminInviteSerializer, EnterpriseCustomerAdminSerializer
from enterprise.constants import (
    ACTIVE_ADMIN_ROLE_TYPE,
    ENTERPRISE_ADMIN_ROLE,
    ENTERPRISE_CUSTOMER_PROVISIONING_ADMIN_ACCESS_PERMISSION,
    PENDING_ADMIN_ROLE_TYPE,
)

logger = logging.getLogger(__name__)

User = get_user_model()


class EnterpriseCustomerAdminPagination(PageNumberPagination):
    """
    Pagination class for EnterpriseCustomerAdmin viewset.
    """
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100


class EnterpriseCustomerAdminViewSet(
    mixins.ListModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
):
    """
    API views for the ``enterprise-customer-admin`` API endpoint.
    Only allows GET, and PATCH requests.
    """
    queryset = models.EnterpriseCustomerAdmin.objects.all()
    serializer_class = EnterpriseCustomerAdminSerializer
    permission_classes = (IsAuthenticated,)
    pagination_class = EnterpriseCustomerAdminPagination

    def get_queryset(self):
        """
        Filter queryset to only show records for the admin user.
        Excludes admins whose EnterpriseCustomerUser has been deactivated.
        """
        return models.EnterpriseCustomerAdmin.objects.filter(
            enterprise_customer_user__user_fk=self.request.user,
            enterprise_customer_user__active=True,
        )

    @action(detail=True, methods=['post'])
    def complete_tour_flow(self, request, pk=None):  # pylint: disable=unused-argument
        """
        Add a completed tour flow to the admin's completed_tour_flows.
        POST /api/v1/enterprise-customer-admin/{pk}/complete_tour_flow/

        Request Arguments:
        - ``flow_uuid``: The request object containing the flow_uuid

        Returns: A response indicating success or failure
        """
        admin = self.get_object()
        flow_uuid = request.data.get('flow_uuid')

        if not flow_uuid:
            return Response(
                {'error': 'flow_uuid is required'},
                status=status.HTTP_400_BAD_REQUEST
            )

        try:
            flow = get_object_or_404(models.OnboardingFlow, uuid=flow_uuid)
            admin.completed_tour_flows.add(flow)

            return Response({
                'status': 'success',
                'message': f'Successfully added tour flow {flow.title} to completed flows'
            })

        except models.OnboardingFlow.DoesNotExist:
            return Response(
                {'error': f'OnboardingFlow with uuid {flow_uuid} does not exist'},
                status=status.HTTP_404_NOT_FOUND
            )

    @permission_required(
        ENTERPRISE_CUSTOMER_PROVISIONING_ADMIN_ACCESS_PERMISSION,
        fn=lambda request, *args, **kwargs: request.data.get('enterprise_customer_uuid'),
    )
    @action(detail=False, methods=['post'])
    def create_admin_by_email(self, request):
        """
        Create a new EnterpriseCustomerAdmin record based on an email address.
        The email address must match an existing user.

        POST /api/v1/enterprise-customer-admin/create_admin_by_email/

        The requesting user must have the ``enterprise_provisioning_admin``
        role to access this endpoint.

        Request Arguments:
        - ``email``: Email address of the user to make an admin
        - ``enterprise_customer_uuid``: UUID of the enterprise customer

        Returns: A response with the created admin record.
        """
        email = request.data.get('email')
        enterprise_customer_uuid = request.data.get('enterprise_customer_uuid')

        if not email:
            return Response(
                {'error': 'email is required'}, status=status.HTTP_400_BAD_REQUEST,
            )

        if not enterprise_customer_uuid:
            return Response(
                {'error': 'enterprise_customer_uuid is required'}, status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            user = User.objects.get(email=email)
        except User.DoesNotExist:
            return Response(
                {'error': f'User with email {email} does not exist'}, status=status.HTTP_404_NOT_FOUND,
            )

        try:
            enterprise_customer = models.EnterpriseCustomer.objects.get(uuid=enterprise_customer_uuid)
        except models.EnterpriseCustomer.DoesNotExist:
            return Response(
                {'error': f'EnterpriseCustomer with uuid {enterprise_customer_uuid} does not exist'},
                status=status.HTTP_404_NOT_FOUND
            )

        # Get or create EnterpriseCustomerUser
        enterprise_customer_user, _ = models.EnterpriseCustomerUser.objects.get_or_create(
            enterprise_customer=enterprise_customer,
            user_fk=user,
            defaults={'user_id': user.id}
        )

        response_status_code = status.HTTP_200_OK
        admin, was_created = models.EnterpriseCustomerAdmin.objects.get_or_create(
            enterprise_customer_user=enterprise_customer_user
        )
        if was_created:
            response_status_code = status.HTTP_201_CREATED

        roles_api.assign_admin_role(
            enterprise_customer_user.user,
            enterprise_customer=enterprise_customer_user.enterprise_customer
        )

        serializer = self.get_serializer(admin)
        return Response(serializer.data, status=response_status_code)

    @permission_required(
        ENTERPRISE_CUSTOMER_PROVISIONING_ADMIN_ACCESS_PERMISSION,
        fn=lambda request, *args, **kwargs: kwargs.get('enterprise_customer_uuid'),
    )
    @action(
        detail=False,
        methods=["post"],
        url_path=r"(?P<enterprise_customer_uuid>[0-9a-fA-F-]+)/admin_invites",
        url_name="admin-invites",
    )
    def invite_admins(self, request, **kwargs):
        """
        Invite new admins to an Enterprise Customer by sending invitation emails.

        POST /enterprise/api/v1/enterprise-customer-admin/{enterprise_customer_uuid}/admin-invites

        Request data must include:

        - emails: list of email addresses to invite
        - enterprise_customer_uuid: UUID of the enterprise customer

        Returns:
            A list of dicts with email and status for each attempted invite.
        """
        enterprise_customer_uuid = kwargs.get("enterprise_customer_uuid")
        if not enterprise_customer_uuid:
            logger.warning("Missing enterprise_customer_uuid in request URL/path.")
            return Response({"detail": "Missing enterprise_customer_uuid."}, status=status.HTTP_400_BAD_REQUEST)

        try:
            enterprise_customer = get_object_or_404(models.EnterpriseCustomer, uuid=enterprise_customer_uuid)
        except ValidationError:
            logger.warning("Invalid enterprise_customer_uuid format: %s", enterprise_customer_uuid)
            return Response(
                {"detail": "enterprise_customer_uuid must be a valid UUID."},
                status=status.HTTP_400_BAD_REQUEST,
            )
        serializer = AdminInviteSerializer(data=request.data)
        if not serializer.is_valid():
            logger.info(
                "Invalid payload for enterprise customer: %s, errors: %s",
                enterprise_customer_uuid,
                serializer.errors
            )
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        normalized_emails = serializer.validated_data.get("emails", [])

        logger.info(
            "Inviting admins for enterprise customer: %s, email count: %d",
            enterprise_customer_uuid,
            len(normalized_emails),
        )

        # Batch prefetch to avoid N+1 queries
        try:
            existing_admin_emails = admin_utils.get_existing_admin_emails(enterprise_customer)
            existing_pending_emails = admin_utils.get_existing_pending_emails(
                enterprise_customer, normalized_emails
            )
        except DatabaseError as exc:
            logger.error(
                "Database error fetching admin status for enterprise %s: %s",
                enterprise_customer_uuid,
                str(exc),
                exc_info=True
            )
            return Response(
                {"detail": "Failed to retrieve admin information due to a database error."},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )

        # Exclude active admins and pending invites from new invites
        new_invites = [
            email for email in normalized_emails
            if email not in existing_admin_emails
            and email not in existing_pending_emails
        ]

        # Track which emails were actually newly created (not just attempted)
        newly_created_emails = set()
        if new_invites:
            logger.info(
                "Creating new pending invites for enterprise customer: %s, new invite count: %d",
                enterprise_customer_uuid,
                len(new_invites),
            )
            try:
                with transaction.atomic():
                    created_invites = admin_utils.create_pending_invites(enterprise_customer, new_invites)
                    # Track emails that were actually created (handles race conditions in get_or_create)
                    newly_created_emails = {invite.user_email for invite in created_invites}
            except DatabaseError as exc:
                logger.error(
                    "Database error creating pending invites for enterprise %s: %s",
                    enterprise_customer_uuid,
                    str(exc),
                    exc_info=True
                )
                return Response(
                    {"detail": "Failed to create pending invites due to a database error."},
                    status=status.HTTP_500_INTERNAL_SERVER_ERROR
                )

        # Update existing_pending_emails with race condition results
        # Emails in new_invites but not in newly_created_emails hit a race condition
        # (get_or_create found an existing record), so add them to existing_pending_emails
        race_condition_emails = set(new_invites) - newly_created_emails
        existing_pending_emails.update(race_condition_emails)

        response_data = []
        for email in normalized_emails:
            status_str = admin_utils.get_invite_status(
                email,
                existing_admin_emails,
                existing_pending_emails
            )
            response_data.append({"email": email, "status": status_str})

        logger.info(
            "Invite response for enterprise customer: %s, total processed: %d, new invites: %d",
            enterprise_customer_uuid,
            len(response_data),
            len(new_invites),
        )
        return Response(response_data, status=status.HTTP_200_OK)

    @permission_required(
        ENTERPRISE_CUSTOMER_PROVISIONING_ADMIN_ACCESS_PERMISSION,
        fn=lambda request, *args, **kwargs: kwargs.get('enterprise_customer_uuid'),
    )
    def delete_admin(self, request, enterprise_customer_uuid=None, id=None):  # pylint: disable=redefined-builtin,unused-argument
        """
        Delete an admin record based on role.

        DELETE /enterprise/api/v1/enterprise-customer/{enterprise_customer_uuid}/admins/{id}/?role=<role>

        Path Parameters:

        - ``enterprise_customer_uuid``: UUID of the enterprise customer.

        - ``id``: ID of the admin record to delete (PendingEnterpriseCustomerAdminUser ID
          or EnterpriseCustomerUser ID)

        Query Parameters:

        - ``role``: Either 'pending' or 'admin' (required, case-insensitive)

        Role handling:
        ``role='pending'`` hard deletes ``PendingEnterpriseCustomerAdminUser`` where ``id=id``.
        ``role='admin'`` deletes role assignment from ``SystemWideEnterpriseUserRoleAssignment``
        for ``EnterpriseCustomerUser`` ``id=id``, and deactivates ``EnterpriseCustomerUser``
        if no other roles exist.

        Returns:
            200 OK with success message and user_deactivated flag (for active admins)
            400 BAD REQUEST if role parameter is missing or invalid
            404 NOT FOUND if the specified admin record doesn't exist
        """
        if not enterprise_customer_uuid:
            logger.warning("Missing enterprise_customer_uuid in request URL/path.")
            return Response({"detail": "Missing enterprise_customer_uuid."}, status=status.HTTP_400_BAD_REQUEST)

        try:
            get_object_or_404(models.EnterpriseCustomer, uuid=enterprise_customer_uuid)
        except ValidationError:
            logger.warning("Invalid enterprise_customer_uuid format: %s", enterprise_customer_uuid)
            return self._error_response(
                'enterprise_customer_uuid must be a valid UUID',
                status.HTTP_400_BAD_REQUEST,
            )

        # Validate id is a valid integer
        try:
            admin_record_id = int(id)
        except (ValueError, TypeError):
            return self._error_response(
                'id must be a valid integer',
                status.HTTP_400_BAD_REQUEST
            )

        role = request.query_params.get('role') or request.data.get('role')

        if not role:
            return self._error_response(
                f'role parameter is required ({PENDING_ADMIN_ROLE_TYPE} or {ACTIVE_ADMIN_ROLE_TYPE})',
                status.HTTP_400_BAD_REQUEST
            )

        role = role.lower()

        if role == PENDING_ADMIN_ROLE_TYPE:
            try:
                return self._delete_pending_admin(enterprise_customer_uuid, admin_record_id)
            except DatabaseError:
                logger.exception(
                    "Database error deleting PendingEnterpriseCustomerAdminUser id=%s for enterprise=%s",
                    admin_record_id,
                    enterprise_customer_uuid,
                )
                return self._error_response(
                    'Failed to delete pending admin invitation due to a database error',
                    status.HTTP_500_INTERNAL_SERVER_ERROR,
                )
        elif role == ACTIVE_ADMIN_ROLE_TYPE:
            try:
                return self._delete_active_admin(enterprise_customer_uuid, admin_record_id)
            except DatabaseError:
                logger.exception(
                    "Database error deleting active admin for EnterpriseCustomerUser id=%s in enterprise=%s",
                    admin_record_id,
                    enterprise_customer_uuid,
                )
                return self._error_response(
                    'Failed to delete admin due to a database error',
                    status.HTTP_500_INTERNAL_SERVER_ERROR,
                )
        else:
            return self._error_response(
                f'Invalid role. Must be "{PENDING_ADMIN_ROLE_TYPE}" or "{ACTIVE_ADMIN_ROLE_TYPE}"',
                status.HTTP_400_BAD_REQUEST
            )

    def _error_response(self, message, status_code):
        """Helper method to create error responses."""
        return Response({'error': message}, status=status_code)

    @transaction.atomic
    def _delete_pending_admin(self, enterprise_customer_uuid, admin_record_id):
        """
        Delete a pending admin invitation.

        Args:
            enterprise_customer_uuid: UUID of the enterprise customer in request path
            admin_record_id: ID of the PendingEnterpriseCustomerAdminUser record

        Returns:
            Response object with success or error message
        """
        try:
            pending_admin = models.PendingEnterpriseCustomerAdminUser.objects.select_for_update().select_related(
                'enterprise_customer'
            ).get(
                id=admin_record_id,
                enterprise_customer__uuid=enterprise_customer_uuid,
            )
            enterprise_customer = pending_admin.enterprise_customer
            user_email = pending_admin.user_email

            pending_admin.delete()
            logger.info(
                "Hard deleted PendingEnterpriseCustomerAdminUser id=%s for enterprise %s",
                admin_record_id,
                enterprise_customer.uuid
            )
            return Response(
                {'message': f'Pending admin invitation for {user_email} deleted successfully'},
                status=status.HTTP_200_OK
            )
        except models.PendingEnterpriseCustomerAdminUser.DoesNotExist:
            return self._error_response(
                'Pending admin invitation not found',
                status.HTTP_404_NOT_FOUND
            )

    @transaction.atomic
    def _delete_active_admin(self, enterprise_customer_uuid, admin_record_id):
        """
        Delete an active admin by removing their role assignment.

        Args:
            enterprise_customer_uuid: UUID of the enterprise customer in request path
            admin_record_id: ID of the EnterpriseCustomerUser record

        Returns:
            Response object with success message and user_deactivated flag
        """
        try:
            enterprise_customer_user = models.EnterpriseCustomerUser.objects.select_for_update().select_related(
                'enterprise_customer', 'user_fk'
            ).get(
                id=admin_record_id,
                active=True,
                enterprise_customer__uuid=enterprise_customer_uuid,
            )
        except models.EnterpriseCustomerUser.DoesNotExist:
            return self._error_response('Admin user not found', status.HTTP_404_NOT_FOUND)

        # Verify and lock admin record so role/admin state transitions stay consistent.
        admin_record = models.EnterpriseCustomerAdmin.objects.select_for_update().filter(
            enterprise_customer_user=enterprise_customer_user
        ).first()
        if not admin_record:
            return self._error_response('Admin record not found', status.HTTP_404_NOT_FOUND)

        enterprise_customer = enterprise_customer_user.enterprise_customer
        user = enterprise_customer_user.user_fk

        # Check if admin role assignment exists
        role_assignment = models.SystemWideEnterpriseUserRoleAssignment.objects.select_for_update().filter(
            user=user,
            enterprise_customer=enterprise_customer,
            role__name=ENTERPRISE_ADMIN_ROLE,
        )

        deleted_count, _ = role_assignment.delete()
        if deleted_count == 0:
            return self._error_response(
                'Admin role assignment not found',
                status.HTTP_404_NOT_FOUND
            )

        logger.info(
            "Deleted %d admin role assignment(s) for user %s in enterprise %s",
            deleted_count,
            user.id,
            enterprise_customer.uuid
        )

        # Check if user has other roles for this enterprise
        # No locking needed - we're only checking existence, and ECU is already locked
        has_other_roles_in_enterprise = models.SystemWideEnterpriseUserRoleAssignment.objects.filter(
            user=user,
            enterprise_customer=enterprise_customer,
        ).exists()

        # Deactivate EnterpriseCustomerUser if no other roles exist in this enterprise (soft delete)
        user_deactivated = False
        if not has_other_roles_in_enterprise:
            # Set both active=False and linked=False to prevent signal from recreating roles
            enterprise_customer_user.active = False
            enterprise_customer_user.linked = False
            enterprise_customer_user.save(update_fields=['active', 'linked', 'modified'])
            logger.info(
                "Deactivated EnterpriseCustomerUser id=%s in enterprise %s (no other roles in this enterprise)",
                enterprise_customer_user.id,
                enterprise_customer.uuid
            )

            # Only deactivate global user account if user has NO access paths remaining
            has_systemwide_roles = models.SystemWideEnterpriseUserRoleAssignment.objects.filter(
                user=user,
            ).exists()

            # Check for other enterprise role constructs that may still grant access
            has_feature_roles = models.EnterpriseFeatureUserRoleAssignment.objects.filter(
                user=user,
            ).exists()

            # Check for Django-level privileged access
            has_django_privileged_access = bool(
                getattr(user, "is_staff", False) or getattr(user, "is_superuser", False)
            )

            has_any_access_paths = has_systemwide_roles or has_feature_roles or has_django_privileged_access

            if not has_any_access_paths:
                user.is_active = False
                user.save(update_fields=['is_active'])
                user_deactivated = True
                logger.info(
                    "Deactivated auth_user id=%s (no enterprise roles or privileged access)",
                    user.id
                )
            else:
                logger.info(
                    "Auth_user id=%s remains active (has enterprise roles or privileged access)",
                    user.id
                )
        else:
            logger.info(
                "EnterpriseCustomerUser id=%s active for user %s (has other roles in this enterprise)",
                enterprise_customer_user.id,
                user.id
            )

        # Create meaningful message with email
        user_identifier = user.email or user.username
        message = (
            f'Admin {user_identifier} deleted successfully and user account deactivated'
            if user_deactivated else
            f'Admin {user_identifier} deleted successfully'
        )

        return Response(
            {
                'message': message,
                'user_deactivated': user_deactivated
            },
            status=status.HTTP_200_OK
        )



# Generated by Django 3.2.22 on 2024-02-06 11:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0032_alter_historicalcanvasenterprisecustomerconfiguration_options'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='canvaslearnerdatatransmissionaudit',
            constraint=models.UniqueConstraint(fields=('enterprise_course_enrollment_id', 'course_id'), name='canvas_unique_enrollment_course_id'),
        ),
    ]



"""
Tests for migrations, especially potentially risky data migrations.
"""
from importlib.metadata import version
from io import StringIO

import pytest

from django.core.management import call_command
from django.test.testcases import TestCase
from django.test.utils import override_settings

GET_DJANGO_VERSION = int(version('django').split('.')[0])


@pytest.mark.skipif(GET_DJANGO_VERSION > 3, reason="django4.2 brings new migrations, so only run for dj32 for now.")
class MigrationTests(TestCase):
    """
    Runs migration tests using Django Command interface.
    """

    @override_settings(MIGRATION_MODULES={})
    def test_migrations_are_in_sync(self):
        out = StringIO()
        call_command("makemigrations", dry_run=True, verbosity=3, stdout=out)
        output = out.getvalue()
        self.assertIn("No changes detected", output)



# Generated by Django 2.2.17 on 2020-12-01 18:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0005_auto_20200909_1534'),
    ]

    operations = [
        migrations.CreateModel(
            name='CanvasLearnerAssessmentDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('canvas_user_email', models.CharField(max_length=255)),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(db_index=True)),
                ('course_id', models.CharField(help_text="The course run's key which is used to uniquely identify the course for Canvas.", max_length=255)),
                ('subsection_id', models.CharField(db_index=True, help_text="The course's subsections's key.", max_length=255)),
                ('grade_point_score', models.FloatField(help_text='The amount of points that the learner scored on the subsection.')),
                ('grade_points_possible', models.FloatField(help_text='The total amount of points that the learner could score on the subsection.')),
                ('grade', models.FloatField(help_text='The grade an enterprise learner received on the reported subsection.')),
                ('subsection_name', models.CharField(help_text="The name given to the subsection being reported. Used for displaying on external LMS'.", max_length=255)),
                ('status', models.CharField(max_length=100)),
                ('error_message', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



# Generated by Django 3.2.20 on 2023-09-06 14:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0182_auto_20230829_1741'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='enable_learner_portal_offers',
            field=models.BooleanField(default=False, help_text='Specifies whether enterprise offers will be made known to learners in the learner portal. This only applies to customers with “offers”, the old version of learner credit.', verbose_name='Enable learner credit in the learner portal'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='enable_learner_portal_offers',
            field=models.BooleanField(default=False, help_text='Specifies whether enterprise offers will be made known to learners in the learner portal. This only applies to customers with “offers”, the old version of learner credit.', verbose_name='Enable learner credit in the learner portal'),
        ),
    ]



# Generated by Django 1.11.23 on 2019-10-01 07:42


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('degreed', '0007_auto_20190925_0730'),
    ]

    operations = [
        migrations.AddField(
            model_name='degreedenterprisecustomerconfiguration',
            name='catalogs_to_transmit',
            field=models.TextField(blank=True, help_text='A comma-separated list of catalog UUIDs to transmit.', null=True),
        ),
        migrations.AddField(
            model_name='historicaldegreedenterprisecustomerconfiguration',
            name='catalogs_to_transmit',
            field=models.TextField(blank=True, help_text='A comma-separated list of catalog UUIDs to transmit.', null=True),
        ),
    ]



"""
Utility functions for the Enterprise API.
"""
import logging
from typing import List, Set

from django.conf import settings
from django.contrib import auth
from django.db import DatabaseError, transaction
from django.db.models import F
from django.db.models.functions import Lower
from django.utils.translation import gettext as _

from enterprise.constants import (
    BRAZE_ADMIN_INVITE_CAMPAIGN_SETTING,
    BRAZE_LEARNER_INVITE_CAMPAIGN_SETTING,
    ENTERPRISE_ADMIN_ROLE,
    ENTERPRISE_CATALOG_ADMIN_ROLE,
    ENTERPRISE_DASHBOARD_ADMIN_ROLE,
    ENTERPRISE_REPORTING_CONFIG_ADMIN_ROLE,
    AdminInviteStatus,
)
from enterprise.models import (
    EnterpriseCustomer,
    EnterpriseCustomerAdmin,
    EnterpriseCustomerCatalog,
    EnterpriseCustomerInviteKey,
    EnterpriseCustomerReportingConfiguration,
    EnterpriseCustomerUser,
    EnterpriseFeatureRole,
    EnterpriseFeatureUserRoleAssignment,
    EnterpriseGroup,
    PendingEnterpriseCustomerAdminUser,
    SystemWideEnterpriseUserRoleAssignment,
)
from enterprise.tasks import send_enterprise_admin_invite_email

logger = logging.getLogger(__name__)


def get_existing_admin_emails(enterprise_customer: EnterpriseCustomer) -> Set[str]:
    """
    Retrieve normalized email addresses of existing ACTIVE enterprise admins.

    Only includes admins who have:
    1. An EnterpriseCustomerAdmin record
    2. An active EnterpriseCustomerUser (active=True)
    3. An active admin role assignment in SystemWideEnterpriseUserRoleAssignment

    Args:
        enterprise_customer: The enterprise customer instance.

    Returns:
        Set of lowercased email addresses of active admins with valid role assignments.

    Raises:
        DatabaseError: If database query fails.

    Example:
        >>> emails = get_existing_admin_emails(customer)
        >>> 'admin@example.com' in emails
        True
    """
    try:
        # Get user IDs with active admin role assignments
        users_with_admin_role = set(
            SystemWideEnterpriseUserRoleAssignment.objects.filter(
                enterprise_customer=enterprise_customer,
                role__name=ENTERPRISE_ADMIN_ROLE,
            ).values_list('user_id', flat=True)
        )

        # Return emails of admins who have active ECU AND active role assignment
        return set(
            EnterpriseCustomerAdmin.objects.filter(
                enterprise_customer_user__enterprise_customer=enterprise_customer,
                enterprise_customer_user__active=True,
                enterprise_customer_user__user_id__in=users_with_admin_role,
            )
            .annotate(email_l=Lower(F("enterprise_customer_user__user_fk__email")))
            .values_list("email_l", flat=True)
        )
    except DatabaseError:
        logger.exception(
            "Database error retrieving existing admin emails for enterprise customer: %s",
            enterprise_customer.uuid,
        )
        raise


def get_existing_pending_emails(
    enterprise_customer: EnterpriseCustomer,
    normalized_emails: List[str]
) -> Set[str]:
    """
    Retrieve normalized email addresses of pending admin invitations.

    Args:
        enterprise_customer: The enterprise customer instance.
        normalized_emails: List of normalized email addresses to check.

    Returns:
        Set of lowercased email addresses that have pending invitations.

    Raises:
        DatabaseError: If database query fails.

    Example:
        >>> pending = get_existing_pending_emails(customer, ['user@example.com'])
        >>> 'user@example.com' in pending
        True
    """
    try:
        return set(
            PendingEnterpriseCustomerAdminUser.objects.filter(
                enterprise_customer=enterprise_customer,
            )
            .annotate(email_l=Lower(F("user_email")))
            .filter(email_l__in=normalized_emails)
            .values_list("email_l", flat=True)
        )
    except DatabaseError:
        logger.exception(
            "Database error retrieving existing pending emails for enterprise customer: %s",
            enterprise_customer.uuid,
        )
        raise


def create_pending_invites(
    enterprise_customer: EnterpriseCustomer,
    emails_to_invite: List[str]
) -> List[PendingEnterpriseCustomerAdminUser]:
    """
    Create pending admin invitations and trigger email notifications.

    Creates PendingEnterpriseCustomerAdminUser records for new admin invites
    and enqueues Braze email tasks to be sent after transaction commits.

    Args:
        enterprise_customer: The enterprise customer instance.
        emails_to_invite: List of normalized email addresses to invite.

    Returns:
        List of created PendingEnterpriseCustomerAdminUser instances.

    Raises:
        DatabaseError: If database operation fails.
        ValueError: If emails_to_invite is empty.
        RuntimeError: If called outside a transaction.atomic block.

    Note:
        - Caller must wrap in transaction.atomic() to ensure atomicity
        - Uses get_or_create per email to avoid duplicate invite emails in race conditions
        - Emails are queued via transaction.on_commit() to send after transaction commits
        - Emails are routed to different Braze campaigns based on EnterpriseCustomerUser existence
          determined at invite creation time (before transaction commits)
        - This ensures emails only send if database changes succeed

    Example:
        >>> with transaction.atomic():
        ...     invites = create_pending_invites(customer, ['new@example.com'])
        >>> len(invites) > 0
        True
    """
    if not emails_to_invite:
        raise ValueError("emails_to_invite cannot be empty")

    if not transaction.get_connection().in_atomic_block:
        raise RuntimeError("create_pending_invites must be called inside transaction.atomic().")

    def _enqueue_email_jobs(customer, invites, ecu_emails):
        """Enqueue invite emails, routing to appropriate Braze campaigns."""
        if not invites:
            return

        try:
            created_invite_emails = [invite.user_email for invite in invites]

            # Split emails based on pre-determined ECU existence
            # Using existing_ecu_emails captured before transaction to avoid race conditions
            learner_emails = []
            new_admin_emails = []
            for email in created_invite_emails:
                if email in ecu_emails:
                    learner_emails.append(email)
                else:
                    new_admin_emails.append(email)

            # Send to existing learners with learner campaign
            if learner_emails:
                send_enterprise_admin_invite_email.delay(
                    str(customer.uuid),
                    learner_emails,
                    campaign_setting_name=BRAZE_LEARNER_INVITE_CAMPAIGN_SETTING
                )

            # Send to new admins with admin campaign
            if new_admin_emails:
                send_enterprise_admin_invite_email.delay(
                    str(customer.uuid),
                    new_admin_emails,
                    campaign_setting_name=BRAZE_ADMIN_INVITE_CAMPAIGN_SETTING
                )

        except Exception:  # pylint: disable=broad-except
            # Log email queueing failures but don't fail the transaction
            # Invites are created successfully, emails can be re-sent manually
            logger.exception(
                "Failed to enqueue admin invite emails for enterprise customer: %s. "
                "Invite count: %d",
                customer.uuid,
                len(invites)
            )

    try:
        # Query existing active EnterpriseCustomerUsers BEFORE creating invites to determine routing
        # This prevents race conditions where ECU is created between invite and email sending
        existing_ecu_emails = set(
            EnterpriseCustomerUser.objects.filter(
                enterprise_customer=enterprise_customer,
                user_id__isnull=False,
                active=True,
            ).select_related('user_fk').annotate(
                email_lower=Lower('user_fk__email')
            ).filter(
                email_lower__in=emails_to_invite
            ).values_list('email_lower', flat=True)
        )

        created_invites = []
        for email in emails_to_invite:
            pending_invite, created = PendingEnterpriseCustomerAdminUser.objects.get_or_create(
                enterprise_customer=enterprise_customer,
                user_email=email,
            )
            if created:
                created_invites.append(pending_invite)

        transaction.on_commit(
            lambda: _enqueue_email_jobs(enterprise_customer, created_invites, existing_ecu_emails)
        )
        return created_invites

    except DatabaseError:
        logger.exception(
            "Database error creating pending invites for enterprise customer: %s",
            enterprise_customer.uuid,
        )
        raise


def get_invite_status(
    email: str,
    existing_admin_emails: Set[str],
    existing_pending_emails: Set[str]
) -> str:
    """
    Determine the invitation status for a given email address.

    Args:
        email (str): The email address to check.
        existing_admin_emails (Set[str]): Set of existing active admin email addresses.
        existing_pending_emails (Set[str]): Set of pending invitation email addresses.

    Returns:
        str: Status constant indicating email state:
            - AdminInviteStatus.EXISTING_ADMIN if user is already an active admin
            - AdminInviteStatus.PENDING_INVITE if invitation already sent
            - AdminInviteStatus.NEW_INVITE if this is a new invitation

    Example:
        >>> status = get_invite_status('new@example.com', set(), set())
        >>> status == 'invite sent'
        True
    """
    if email in existing_admin_emails:
        return AdminInviteStatus.EXISTING_ADMIN
    if email in existing_pending_emails:
        return AdminInviteStatus.PENDING_INVITE
    return AdminInviteStatus.NEW_INVITE


User = auth.get_user_model()
SERVICE_USERNAMES = (
    'ECOMMERCE_SERVICE_WORKER_USERNAME',
    'ENTERPRISE_SERVICE_WORKER_USERNAME'
)


class CourseRunProgressStatuses:
    """
    Class to group statuses that a course run can be in with respect to user progress.
    """

    IN_PROGRESS = 'in_progress'
    UPCOMING = 'upcoming'
    COMPLETED = 'completed'
    SAVED_FOR_LATER = 'saved_for_later'


def get_service_usernames():
    """
    Return the set of service usernames that are given extended permissions in the API.
    """
    if getattr(settings, 'ENTERPRISE_ALL_SERVICE_USERNAMES', None):
        return set(settings.ENTERPRISE_ALL_SERVICE_USERNAMES)

    return {getattr(settings, username, None) for username in SERVICE_USERNAMES}


def get_enterprise_customer_from_enterprise_group_id(group_id):
    """
    Get the enterprise customer id given an enterprise customer group id.
    """
    try:
        return str(EnterpriseGroup.objects.get(pk=group_id).enterprise_customer.uuid)
    except EnterpriseGroup.DoesNotExist:
        return None


def get_enterprise_customer_from_catalog_id(catalog_id):
    """
    Get the enterprise customer id given an enterprise customer catalog id.
    """
    try:
        return str(EnterpriseCustomerCatalog.objects.get(pk=catalog_id).enterprise_customer.uuid)
    except EnterpriseCustomerCatalog.DoesNotExist:
        return None


def get_ent_cust_from_report_config_uuid(uuid):
    """
    Get the enterprise customer id given an enterprise report configuration UUID.
    """
    try:
        return str(EnterpriseCustomerReportingConfiguration.objects.get(uuid=uuid).enterprise_customer.uuid)
    except EnterpriseCustomerReportingConfiguration.DoesNotExist:
        return None


def get_enterprise_customer_from_user_id(user_id):
    """
    Get the enterprise customer id given an user id
    """
    try:
        return str(EnterpriseCustomerUser.objects.get(user_id=user_id).enterprise_customer.uuid)
    except EnterpriseCustomerUser.DoesNotExist:
        return None


def create_message_body(email, enterprise_name, number_of_codes=None, notes=None):
    """
    Return the message body with extra information added by user.
    """
    if number_of_codes and notes:
        body_msg = _('{token_email} from {token_enterprise_name} has requested {token_number_codes} additional '
                     'codes. Please reach out to them.\nAdditional Notes:\n{token_notes}.').format(
                         token_email=email,
                         token_enterprise_name=enterprise_name,
                         token_number_codes=number_of_codes,
                         token_notes=notes)
    elif number_of_codes:
        body_msg = _('{token_email} from {token_enterprise_name} has requested {token_number_codes} additional '
                     'codes. Please reach out to them.').format(
                         token_email=email,
                         token_enterprise_name=enterprise_name,
                         token_number_codes=number_of_codes)
    elif notes:
        body_msg = _('{token_email} from {token_enterprise_name} has requested additional '
                     'codes. Please reach out to them.\nAdditional Notes:\n{token_notes}.').format(
                         token_email=email,
                         token_enterprise_name=enterprise_name,
                         token_notes=notes)
    else:
        body_msg = _('{token_email} from {token_enterprise_name} has requested additional codes.'
                     ' Please reach out to them.').format(
                         token_email=email,
                         token_enterprise_name=enterprise_name)
    return body_msg


def get_ent_cust_from_enterprise_customer_key(enterprise_customer_key):
    """
    Get the enterprise customer id given an enterprise customer key.
    """

    try:
        return str(EnterpriseCustomerInviteKey.objects.get(uuid=enterprise_customer_key).enterprise_customer_id)
    except EnterpriseCustomerInviteKey.DoesNotExist:
        return None


def delta_format(current, prior):
    """
    Formate delta of the given numbers.

    If the delta is positive, number is '+10'. If negative, change nothing, it will come through as '-10' by default.
    """
    delta = current - prior
    if delta >= 0:
        return f'+{delta}'
    else:
        return f'{delta}'


def percentage_format(number):
    """
    Turn float representation of percentage into a cleaner format (0.89 -> 89%)

    Arguments:
        number (float): Floating point number to format.

    Returns:
        (str): String representation of the float in percentage form.
    """
    return f'{number * 100:.0f}%'


def generate_prompt_for_learner_engagement_summary(engagement_data):
    """
    Generate an OpenAI prompt to get the summary of learner engagement from engagement data.

    Arguments:
         engagement_data (dict): A dictionary containing learner engagement numbers for some enterprise customer.

    Returns:
        (str): OpenAI prompt for getting engagement summary.
    """
    is_active = engagement_data['active_contract']
    data = {
        'enrolls': engagement_data['enrolls'],
        'enrolls_delta': delta_format(current=engagement_data['enrolls'], prior=engagement_data['enrolls_prior']),
        'engage': engagement_data['engage'],
        'engage_delta': delta_format(current=engagement_data['engage'], prior=engagement_data['engage_prior']),
        'hours': engagement_data['hours'],
        'hours_delta': delta_format(current=engagement_data['hours'], prior=engagement_data['hours_prior']),
        'passed': engagement_data['passed'],
        'passed_delta': delta_format(current=engagement_data['passed'], prior=engagement_data['passed_prior']),
    }

    # If active contract (or unknown).
    if is_active or is_active is None:
        prompt = settings.LEARNER_ENGAGEMENT_PROMPT_FOR_ACTIVE_CONTRACT
    else:
        # if enterprise customer does not have an active contract
        prompt = settings.LEARNER_ENGAGEMENT_PROMPT_FOR_NON_ACTIVE_CONTRACT

    return prompt.format(**data)


def generate_prompt_for_learner_progress_summary(progress_data):
    """
    Generate an OpenAI prompt to get the summary of learner progress from progress data.

    Arguments:
         progress_data (dict): A dictionary containing learner progress numbers for some enterprise customer.

    Returns:
        (str): OpenAI prompt for getting learner progress summary.
    """
    is_active = progress_data['active_subscription_plan']
    data = {
        'assigned_licenses': progress_data['assigned_licenses'],
        'assigned_licenses_percentage': percentage_format(progress_data['assigned_licenses_percentage']),
        'activated_licenses': progress_data['activated_licenses'],
        'activated_licenses_percentage': percentage_format(progress_data['activated_licenses_percentage']),
        'active_enrollments': progress_data['active_enrollments'],
        'at_risk_enrollment_less_than_one_hour': progress_data['at_risk_enrollment_less_than_one_hour'],
        'at_risk_enrollment_end_date_soon': progress_data['at_risk_enrollment_end_date_soon'],
        'at_risk_enrollment_dormant': progress_data['at_risk_enrollment_dormant'],
    }

    # If active contract (or unknown).
    if is_active or is_active is None:
        prompt = settings.LEARNER_PROGRESS_PROMPT_FOR_ACTIVE_CONTRACT
    else:
        # if enterprise customer does not have an active contract
        prompt = settings.LEARNER_PROGRESS_PROMPT_FOR_NON_ACTIVE_CONTRACT

    return prompt.format(**data)


def set_application_name_from_user_id(user_id):
    """
    Get the enterprise customer user's name given a user id.
    """
    try:
        user = User.objects.get(id=user_id)
        return f"{user.username}'s Enterprise Credentials"
    except User.DoesNotExist:
        return None


def has_api_credentials_enabled(enterprise_uuid):
    """
    Check whether the enterprise customer can access to api credentials or not
    """
    try:
        return (EnterpriseCustomer
                .objects.get(uuid=enterprise_uuid)
                .enable_generation_of_api_credentials)
    except EnterpriseCustomer.DoesNotExist:
        return False


def assign_feature_roles(user):
    """
    Add the ENTERPRISE_DASHBOARD_ADMIN_ROLE, ENTERPRISE_REPORTING_CONFIG_ADMIN_ROLE, ENTERPRISE_CATALOG_ADMIN_ROLE
    feature roles if the user does not already have them
    """
    roles_name = [ENTERPRISE_DASHBOARD_ADMIN_ROLE, ENTERPRISE_REPORTING_CONFIG_ADMIN_ROLE,
                  ENTERPRISE_CATALOG_ADMIN_ROLE]
    for role_name in roles_name:
        feature_role_object, __ = EnterpriseFeatureRole.objects.get_or_create(name=role_name)
        EnterpriseFeatureUserRoleAssignment.objects.get_or_create(user=user, role=feature_role_object)



"""
Utility functions for the SAP SuccessFactors integrated channel.
"""

from logging import getLogger

from integrated_channels.sap_success_factors.constants import SUCCESSFACTORS_OCN_LANGUAGE_CODES

LOGGER = getLogger(__name__)


def transform_language_code(code):
    """
    Transform ISO language code (e.g. en-us) to the language name expected by SAPSF.
    """
    if code is None:
        return 'English'

    components = code.split('-', 2)
    language_code = components[0]
    try:
        country_code = components[1]
    except IndexError:
        country_code = '_'

    language_family = SUCCESSFACTORS_OCN_LANGUAGE_CODES.get(language_code)
    if not language_family:
        return 'English'

    return language_family.get(country_code, language_family['_'])


def is_sso_enabled_via_tpa_service(enterprise_customer):
    """
    Determines if SSO is enabled via a third party auth service (like Auth0) for customer leveraging
    SAP integration, by checking presence of org_id and active EnterpriseCustomerSsoConfiguration record.
    """
    if enterprise_customer.auth_org_id is None:
        return False

    enterprise_orchestration_config = enterprise_customer.sso_orchestration_records.filter(
        active=True
    )
    if enterprise_orchestration_config.exists() and enterprise_orchestration_config.first().validated_at is not None:
        return True
    return False



# Generated by Django 4.2.16 on 2024-10-18 15:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0223_default_enrollments'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisegroup',
            name='applies_to_all_contexts',
            field=models.BooleanField(default=False, help_text='When enabled, all learners connected to the org will be considered a member.', null=True, verbose_name='Set group membership to the entire org of learners.'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisegroup',
            name='applies_to_all_contexts',
            field=models.BooleanField(default=False, help_text='When enabled, all learners connected to the org will be considered a member.', null=True, verbose_name='Set group membership to the entire org of learners.'),
        ),
    ]



# Generated by Django 4.2.16 on 2025-02-19 20:20

from django.db import migrations, connection


class Migration(migrations.Migration):
    dependencies = [
        ('enterprise', '0229_enterprisecustomeruser_user_fk_and_more'),
    ]

    db_engine = connection.settings_dict['ENGINE']
    if 'mysql' in db_engine:
        operations = [
            # Custom SQL to add indexes with zero-downtime options
            migrations.RunSQL(
                sql="""
                    ALTER TABLE enterprise_enterprisecustomeruser
                        ADD INDEX idx_enterprise_user_fk (user_fk),
                        ALGORITHM = INPLACE,
                        LOCK = NONE;

                    ALTER TABLE enterprise_historicalenterprisecustomeruser
                        ADD INDEX idx_historical_enterprise_user_fk (user_fk),
                        ALGORITHM = INPLACE,
                        LOCK = NONE;
                    """,
                reverse_sql="""
                            ALTER TABLE enterprise_enterprisecustomeruser
                                DROP INDEX idx_enterprise_user_fk;
                            ALTER TABLE enterprise_historicalenterprisecustomeruser
                                DROP INDEX idx_historical_enterprise_user_fk;
                            """
            ),
        ]
    if 'postgres' in db_engine:
        operations = [
            # Custom SQL to add indexes with zero-downtime options
            migrations.RunSQL(
                sql="""
                    CREATE INDEX idx_enterprise_user_fk ON enterprise_enterprisecustomeruser (user_fk);
                    CREATE INDEX idx_historical_enterprise_user_fk ON enterprise_historicalenterprisecustomeruser (user_fk);
                    """,
                reverse_sql="""
                            DROP INDEX idx_enterprise_user_fk;
                            DROP INDEX idx_historical_enterprise_user_fk;
                            """
            ),
        ]
    else:
        operations = []



# Generated by Django 3.2.12 on 2022-06-14 16:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0158_remove_is_active_from_role_assignment'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_learner_portal_offers',
            field=models.BooleanField(default=False, help_text='Specifies whether enterprise offers will be made known to learners in the learner portal.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_learner_portal_offers',
            field=models.BooleanField(default=False, help_text='Specifies whether enterprise offers will be made known to learners in the learner portal.'),
        ),
    ]




"""
Mark already transmitted LearnerDataTransmission as is_trasmitted=True for all integrated channels
"""

from logging import getLogger

from django.apps import apps
from django.core.management.base import BaseCommand

from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin

LOGGER = getLogger(__name__)


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Mark already transmitted LearnerDataTransmission as is_trasmitted=True for all integrated channels
    """

    def handle(self, *args, **options):
        """
        Mark already transmitted LearnerDataTransmission as is_trasmitted=True
        """
        channel_learner_audit_models = [
            ('moodle', 'MoodleLearnerDataTransmissionAudit'),
            ('blackboard', 'BlackboardLearnerDataTransmissionAudit'),
            ('blackboard', 'BlackboardLearnerAssessmentDataTransmissionAudit'),
            ('canvas', 'CanvasLearnerDataTransmissionAudit'),
            ('degreed2', 'Degreed2LearnerDataTransmissionAudit'),
            ('degreed', 'DegreedLearnerDataTransmissionAudit'),
            ('sap_success_factors', 'SapSuccessFactorsLearnerDataTransmissionAudit'),
            ('cornerstone', 'CornerstoneLearnerDataTransmissionAudit'),
            ('canvas', 'CanvasLearnerAssessmentDataTransmissionAudit'),
        ]
        for app_label, model_name in channel_learner_audit_models:
            model_class = apps.get_model(app_label=app_label, model_name=model_name)
            LOGGER.info(
                f'Started: setting {model_name}.is_transmitted set to True'
            )
            model_class.objects.filter(
                error_message='',
                status__lt=400,
            ).update(is_transmitted=True)

            LOGGER.info(
                f'Finished: setting {model_name}.is_transmitted set to True'
            )



"""
Django signal handlers for the consent module.
"""
from logging import getLogger

from consent.models import DataSharingConsent

logger = getLogger(__name__)

try:
    from openedx.core.djangoapps.user_api.accounts.signals import USER_RETIRE_LMS_CRITICAL
except ImportError:
    USER_RETIRE_LMS_CRITICAL = None


def retire_users_data_sharing_consent(sender, user, retired_username, **kwargs):  # pylint: disable=unused-argument
    """
    Handle USER_RETIRE_LMS_CRITICAL signal: retire DataSharingConsent username records.

    Idempotent: only updates records where username hasn't already been changed to the retired value.
    """
    if user.username != retired_username:
        logger.info(
            "Updating username %s to retired username %s on DataSharingConsent records",
            user.username,
            retired_username,
        )
    DataSharingConsent.objects.filter(
        username=user.username,
    ).exclude(
        username=retired_username,
    ).update(username=retired_username)


if USER_RETIRE_LMS_CRITICAL is not None:
    USER_RETIRE_LMS_CRITICAL.connect(retire_users_data_sharing_consent)



"""
    Serializer for Moodle configuration.
"""
from rest_framework import serializers

from integrated_channels.api.serializers import EnterpriseCustomerPluginConfigSerializer
from integrated_channels.moodle.models import MoodleEnterpriseCustomerConfiguration


class MoodleConfigSerializer(EnterpriseCustomerPluginConfigSerializer):
    """
    Serializer for Moodle Enterprise Customer Configuration.
    Handles token or username/password authentication methods.
    """
    class Meta:
        model = MoodleEnterpriseCustomerConfiguration
        extra_fields = (
            'moodle_base_url',
            'service_short_name',
            'category_id',
            'username',
            'password',
            'token',
        )
        fields = EnterpriseCustomerPluginConfigSerializer.Meta.fields + extra_fields

    username = serializers.CharField(required=False, allow_blank=True, write_only=True)
    password = serializers.CharField(required=False, allow_blank=True, write_only=True)
    token = serializers.CharField(required=False, allow_blank=True, write_only=True)

    def _handle_credentials(self, instance, username=None, password=None, token=None):
        """Helper method to handle credential updates."""
        if token is not None:
            instance.decrypted_token = token
            instance.decrypted_username = None
            instance.decrypted_password = None
        elif username is not None and password is not None:
            instance.decrypted_token = None
            instance.decrypted_username = username
            instance.decrypted_password = password

    def create(self, validated_data):
        username = validated_data.pop('username', None)
        password = validated_data.pop('password', None)
        token = validated_data.pop('token', None)

        instance = super().create(validated_data)
        self._handle_credentials(instance, username, password, token)
        instance.save()
        return instance

    def update(self, instance, validated_data):
        username = validated_data.pop('username', None)
        password = validated_data.pop('password', None)
        token = validated_data.pop('token', None)

        instance = super().update(instance, validated_data)
        self._handle_credentials(instance, username, password, token)
        instance.save()
        return instance
